# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import CreateJob, Counter, CutoffDistance, ExternalHamiltonian, \
    InitialPositions, RemoveJob, RandomVelocity, SphereReflection, VerletPositionUpdate, \
    VerletVelocityUpdate, Zeros
from pyiron_contrib.protocol.primitive.two_state import IsGEq, ModIsZero
from pyiron_contrib.protocol.primitive.fts_vertices import CentroidsRunningAverageMix, CentroidsReparameterization, \
    CentroidsSmoothing, PositionsRunningAverage, StringRecenter, StringReflect
from pyiron_contrib.protocol.list import SerialList, ParallelList
from pyiron_contrib.protocol.utils import Pointer

"""
Protocols for the finite temperature string (FTS) method.
"""

__author__ = "Raynol Dsouza, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "23 July, 2019"


class FTSEvolution(CompoundVertex):
    """
    A serial Finite Temperature String (FTS) protocol to compute migration barriers between two stable system.

    NOTE: 1. This protocol is as of now untested with DFT-type reference jobs, and only works for sure, with
        Lammps-type reference jobs.
          2. Convergence criterion is NOT implemented for this protocol, because it runs serially, and would take
        a VERY long time to achieve a good convergence.

    Input attributes:
        TODO: add a vertex to check if all the necessary inputs are provided.
        ref_job_full_path (string): The path to the saved reference job to use for calculating forces and energies.
        structure_initial (Atoms): The initial structure.
        structure_initial (Atoms): The final structure.
        temperature (float): Temperature to run at in K.
        n_steps (int): How many MD steps to run for. (Default is 100.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is None, which runs NVE.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)
        time_step (float): MD time step in fs. (Default is 1.)
        sampling_period (int): Account output every `sampling_period' for the TI operations. (Default is 1, account
            for every MD step.
        thermalization_steps (int): Number of steps the system is thermalized for to reach equilibrium. (Default is
            10 steps.)
        n_images (int): Number of centroids / images. (Default is 5.)
        initial_positions (list/np.ndarray): The initial positions of the images (preferably from NEB).
            Default is None)
        cutoff_factor (float): The cutoff is obtained by taking the first nearest neighbor distance and multiplying
            it by the cutoff factor. A default value of 0.4 is chosen, because taking a cutoff factor of ~0.5
            sometimes let certain reflections off the hook, and we do not want that to happen. (Default is 0.4.)
        mixing_fraction (float): How much of the images' running average of position to mix into the centroid positions
            each time the mixer is called. (Default is 0.1.)
        relax_endpoints (bool): Whether or not to relax the endpoints of the string. (Default is False.)
        smooth_style (string): Apply 'global' or 'local' smoothing. 'global' smoothing considers an array of n_images
            terms while applying smoothing to each image, while 'local' smoothing only considers the left and right
            neighbor of that image. (Default is 'global'.)
        nominal_smoothing (float): How much smoothing to apply to the updating centroid positions (endpoints are
            not effected). The actual smoothing is the product of this nominal value, the number of images, and the
            mixing fraction, ala Vanden-Eijnden and Venturoli (2009). (Default is 0.1.)
        use_reflection (boolean): Turn on or off `SphereReflection`. Using sphere reflection restricts each atom
            in the simulation cell to evolve within the Wigner-Seitz cell of its reference position. This is
            helpful to restrict atom-hopping in the presence of a vacancy at higher temperatures. (Default is True.)

    Output attributes:
        energy_pot (list[float]): Total potential energy of the system in eV.
        positions (list[numpy.ndarray]): Atomic positions in angstroms for each centroid.
        forces (list[numpy.ndarray]): Atomic forces in eV/angstrom for each centroid.
    """

    def __init__(self, **kwargs):
        super(FTSEvolution, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.temperature = 1.
        id_.n_steps = 100
        id_.temperature_damping_timescale = 100.
        id_.overheat_fraction = 2.
        id_.time_step = 1.
        id_.sampling_period = 1
        id_.thermalization_steps = 10
        id_.n_images = 5
        id_.initial_positions = None
        id_.cutoff_factor = 0.45
        id_.mixing_fraction = 0.1
        id_.relax_endpoints = False
        id_.smooth_style = 'global'
        id_.nominal_smoothing = 0.1
        id_.use_reflection = True
        id_._divisor = 1
        id_._total_steps = 0
        id_._project_path = None
        id_._job_name = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initialize_images = CreateJob()
        g.initialize_centroids = CreateJob()
        g.initial_positions = InitialPositions()
        g.initial_velocities = SerialList(RandomVelocity)
        g.initial_forces = Zeros()
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.reflect_string = SerialList(StringReflect)
        g.reflect_atoms = SerialList(SphereReflection)
        g.calc_static_images = SerialList(ExternalHamiltonian)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.running_average_pos = SerialList(PositionsRunningAverage)
        g.check_sampling_period = ModIsZero()
        g.mix = CentroidsRunningAverageMix()
        g.smooth = CentroidsSmoothing()
        g.reparameterize = CentroidsReparameterization()
        g.calc_static_centroids = SerialList(ExternalHamiltonian)
        g.recenter = SerialList(StringRecenter)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initialize_images,
            g.initialize_centroids,
            g.initial_positions,
            g.initial_velocities,
            g.initial_forces,
            g.cutoff,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect_string,  # Comes before atomic reflecting so we can actually trigger a full string reflection!
            g.reflect_atoms,  # Comes after, since even if the string doesn't reflect, an atom might have migrated.
            g.calc_static_images,
            g.verlet_velocities,
            g.running_average_pos,
            g.check_sampling_period, 'true',
            g.mix,
            g.smooth,
            g.reparameterize,
            g.calc_static_centroids,
            g.recenter,
            g.clock,
            g.check_steps
        )
        g.make_edge(g.check_sampling_period, g.recenter, 'false')
        g.starting_vertex = g.initialize_images
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initialize_images
        g.initialize_images.input.n_images = ip.n_images
        g.initialize_images.input.ref_job_full_path = ip.ref_job_full_path
        g.initialize_images.input.structure = ip.structure_initial

        # initialize_centroids
        g.initialize_centroids.input.n_images = ip.n_images
        g.initialize_centroids.input.ref_job_full_path = ip.ref_job_full_path
        g.initialize_centroids.input.structure = ip.structure_initial

        # initial_positions
        g.initial_positions.input.structure_initial = ip.structure_initial
        g.initial_positions.input.structure_final = ip.structure_final
        g.initial_positions.input.initial_positions = ip.initial_positions
        g.initial_positions.input.n_images = ip.n_images

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_images
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure_initial.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure_initial
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # initial_forces
        g.initial_forces.input.shape = ip.structure_initial.positions.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_images
        g.verlet_positions.broadcast.default.positions = gp.initial_positions.output.initial_positions[-1]
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.recenter.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.recenter.output.forces[-1]
        g.verlet_positions.direct.masses = ip.structure_initial.get_masses
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect_string
        g.reflect_string.input.n_children = ip.n_images
        g.reflect_string.direct.default.all_centroid_positions = \
            gp.initial_positions.output.initial_positions[-1]
        g.reflect_string.broadcast.default.centroid_positions = \
            gp.initial_positions.output.initial_positions[-1]
        g.reflect_string.broadcast.default.previous_positions = \
            gp.initial_positions.output.initial_positions[-1]
        g.reflect_string.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_string.direct.all_centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.reflect_string.broadcast.centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.reflect_string.broadcast.previous_positions = gp.recenter.output.positions[-1]
        g.reflect_string.broadcast.previous_velocities = gp.reflect_atoms.output.velocities[-1]
        g.reflect_string.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_string.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect_string.direct.structure = ip.structure_initial

        # reflect_atoms
        g.reflect_atoms.input.n_children = ip.n_images
        g.reflect_atoms.broadcast.default.reference_positions = \
            gp.initial_positions.output.initial_positions[-1]
        g.reflect_atoms.broadcast.default.previous_positions = \
            gp.initial_positions.output.initial_positions[-1]
        g.reflect_atoms.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]
        g.reflect_atoms.direct.default.total_steps = ip.total_steps

        g.reflect_atoms.broadcast.reference_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.reflect_atoms.broadcast.positions = gp.reflect_string.output.positions[-1]
        g.reflect_atoms.broadcast.velocities = gp.reflect_string.output.velocities[-1]
        g.reflect_atoms.broadcast.previous_positions = gp.recenter.output.positions[-1]
        g.reflect_atoms.broadcast.previous_velocities = gp.reflect_atoms.output.velocities[-1]
        g.reflect_atoms.direct.structure = ip.structure_initial
        g.reflect_atoms.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]
        g.reflect_atoms.direct.use_reflection = ip.use_reflection
        g.reflect_atoms.broadcast.total_steps = gp.reflect_atoms.output.total_steps[-1]

        # calc_static_images
        g.calc_static_images.input.n_children = ip.n_images
        g.calc_static_images.direct.structure = ip.structure_initial
        g.calc_static_images.broadcast.project_path = gp.initialize_images.output.project_path[-1]
        g.calc_static_images.broadcast.job_name = gp.initialize_images.output.job_names[-1]
        g.calc_static_images.broadcast.positions = gp.reflect_atoms.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_images
        g.verlet_velocities.broadcast.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.calc_static_images.output.forces[-1]
        g.verlet_velocities.direct.masses = ip.structure_initial.get_masses
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        # running_average_positions
        g.running_average_pos.input.n_children = ip.n_images
        g.running_average_pos.direct.default.thermalization_steps = ip.thermalization_steps
        g.running_average_pos.direct.default.total_steps = ip.total_steps
        g.running_average_pos.direct.default.divisor = ip.divisor
        g.running_average_pos.broadcast.default.running_average_positions = \
            gp.initial_positions.output.initial_positions[-1]

        g.running_average_pos.broadcast.total_steps = gp.running_average_pos.output.total_steps[-1]
        g.running_average_pos.broadcast.divisor = gp.running_average_pos.output.divisor[-1]
        g.running_average_pos.broadcast.running_average_positions = \
            gp.running_average_pos.output.running_average_positions[-1]
        g.running_average_pos.broadcast.positions = gp.reflect_atoms.output.positions[-1]
        g.running_average_pos.direct.structure = ip.structure_initial

        # check_sampling_period
        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = ip.sampling_period

        # mix
        g.mix.input.default.centroids_pos_list = gp.initial_positions.output.initial_positions[-1]
        g.mix.input.centroids_pos_list = gp.reparameterize.output.centroids_pos_list[-1]
        g.mix.input.mixing_fraction = ip.mixing_fraction
        g.mix.input.relax_endpoints = ip.relax_endpoints
        g.mix.input.running_average_positions = gp.running_average_pos.output.running_average_positions[-1]
        g.mix.input.structure = ip.structure_initial

        # smooth
        g.smooth.input.kappa = ip.nominal_smoothing
        g.smooth.input.dtau = ip.mixing_fraction
        g.smooth.input.structure = ip.structure_initial
        g.smooth.input.smooth_style = ip.smooth_style
        g.smooth.input.centroids_pos_list = gp.mix.output.centroids_pos_list[-1]

        # reparameterize
        g.reparameterize.input.centroids_pos_list = gp.smooth.output.centroids_pos_list[-1]
        g.reparameterize.input.structure= ip.structure_initial

        # calc_static_centroids
        g.calc_static_centroids.input.n_children = ip.n_images
        g.calc_static_centroids.direct.structure = ip.structure_initial
        g.calc_static_centroids.broadcast.project_path = gp.initialize_centroids.output.project_path[-1]
        g.calc_static_centroids.broadcast.job_name = gp.initialize_centroids.output.job_names[-1]
        g.calc_static_centroids.broadcast.positions = gp.reparameterize.output.centroids_pos_list[-1]

        # recenter
        g.recenter.input.n_children = ip.n_images
        g.recenter.direct.default.all_centroid_positions = gp.initial_positions.output.initial_positions[-1]
        g.recenter.broadcast.default.centroid_positions = gp.initial_positions.output.initial_positions[-1]
        g.recenter.direct.default.centroid_forces = gp.initial_forces.output.zeros[-1]

        g.recenter.direct.all_centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.recenter.broadcast.centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.recenter.broadcast.centroid_forces = gp.calc_static_centroids.output.forces[-1]
        g.recenter.broadcast.positions = gp.reflect_atoms.output.positions[-1]
        g.recenter.broadcast.forces = gp.calc_static_images.output.forces[-1]
        g.recenter.direct.structure= ip.structure_initial

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static_centroids.output.energy_pot[-1],
            'positions': ~gp.reparameterize.output.centroids_pos_list[-1],
            'forces': ~gp.calc_static_centroids.output.forces[-1]
        }

    def _get_energies(self, frame=None):
        if frame is None:
            return self.graph.calc_static_centroids.output.energy_pot[-1]
        else:
            return self.graph.calc_static_centroids.archive.output.energy_pot.data[frame]

    def plot_string(self, ax=None, frame=None, plot_kwargs=None):
        """
        Plot the string at an input frame. Here, frame is a dump of a step in the run. If `fts_job´ is the name
            of the fts job, the number of dumps can be specified by the user while submitting the job, as:

        >>> fts_job.set_output_whitelist(**{'calc_static_centroids': {'energy_pot': 20}})

        and run the job. Here, it dumps (or records a frame) of `energy_pot´ from the `calc_static_centroids´ vertex
            once every 20 steps.

        Default is plot the string at the final frame, as only the final dump is recorded. (unless specified
            otherwise by the user!)
        """
        if ax is None:
            _, ax = plt.subplots()
        if plot_kwargs is None:
            plot_kwargs = {}
        if 'marker' not in plot_kwargs.keys():
            plot_kwargs = {'marker': 'o'}
        energies = np.array(self._get_energies(frame=frame))
        ax.plot(energies - energies[0], **plot_kwargs)
        ax.set_ylabel("Energy")
        ax.set_xlabel("Centroid")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        return ax

    def _get_directional_barrier(self, frame=None, anchor_element=0, use_minima=False):
        energies = self._get_energies(frame=frame)
        if use_minima:
            reference = energies.min()
        else:
            reference = energies[anchor_element]
        return energies.max() - reference

    def get_forward_barrier(self, frame=None, use_minima=False):
        """
        Get the energy barrier from the 0th image to the highest energy (saddle state).

        Args:
            frame (int): A particular dump. (Default is None, the final dump.)
            use_minima (bool): Whether to use the minima of the energies to compute tha barrier. (Default is
                False, use the 0th value.)

        Returns:
            (float): the forward migration barrier.
        """
        return self._get_directional_barrier(frame=frame, use_minima=use_minima)

    def get_reverse_barrier(self, frame=None, use_minima=False):
        """
        Get the energy barrier from the final image to the highest energy (saddle state).

        Args:
            frame (int): A particular dump. (Default is None, the final dump.)
            use_minima (bool): Whether to use the minima of the energies to compute tha barrier. (Default is
                False, use the nth value.)

        Returns:
            (float): the backward migration barrier.
        """
        return self._get_directional_barrier(frame=frame, anchor_element=-1, use_minima=use_minima)

    def get_barrier(self, frame=None, use_minima=True):
        return self.get_forward_barrier(frame=frame, use_minima=use_minima)
    get_barrier.__doc__ = get_forward_barrier.__doc__


class ProtoFTSEvoSer(Protocol, FTSEvolution):
    pass


class _ConstrainedMD(CompoundVertex):
    """
    A sub-protocol for FTSEvolutionParallel for the evolution of each image. The MD is 'Constrained' as the atoms in
        the cells of each image evolve with the following constrains:

            a. If an image gets closer to a centroid which is NOT its parent centroid, reverse the velocities of
            all the atoms in that image, such that the image remains in the Voronoi cell of the parent centroid.
            b. If any atom in an image moves out of its Wigner-Seitz cell, reverse the velocities of all the atoms
            in that image.

    This sub-protocol is executed in parallel over multiple cores using ParallelList.

    NOTE: Presently for use only with Finite Temperature String (FTS) related protocols.
    """

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.check_steps = IsGEq()
        g.verlet_positions = VerletPositionUpdate()
        g.reflect_string = StringReflect()
        g.reflect_atoms = SphereReflection()
        g.calc_static = ExternalHamiltonian()
        g.verlet_velocities = VerletVelocityUpdate()
        g.running_average_pos = PositionsRunningAverage()
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect_string,
            g.reflect_atoms,
            g.calc_static,
            g.verlet_velocities,
            g.running_average_pos,
            g.clock,
            g.check_steps,
        )
        g.starting_vertex = g.check_steps
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.default.positions = ip.positions
        g.verlet_positions.input.default.velocities = ip.velocities
        g.verlet_positions.input.default.forces = ip.forces

        g.verlet_positions.input.positions = gp.reflect_atoms.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect_string
        g.reflect_string.input.default.previous_positions = ip.positions
        g.reflect_string.input.default.previous_velocities = ip.velocities

        g.reflect_string.input.all_centroid_positions = ip.all_centroid_positions
        g.reflect_string.input.centroid_positions = ip.centroid_positions
        g.reflect_string.input.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_string.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect_string.input.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_string.input.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect_string.input.pbc = ip.structure.pbc
        g.reflect_string.input.cell = ip.structure.cell.array

        # reflect_atoms
        g.reflect_atoms.input.default.previous_positions = ip.positions
        g.reflect_atoms.input.default.previous_velocities = ip.velocities
        g.reflect_atoms.input.default.total_steps = ip.total_steps

        g.reflect_atoms.input.reference_positions = ip.centroid_positions
        g.reflect_atoms.input.positions = gp.reflect_string.output.positions[-1]
        g.reflect_atoms.input.velocities = gp.reflect_string.output.velocities[-1]
        g.reflect_atoms.input.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_atoms.input.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect_atoms.input.structure = ip.structure
        g.reflect_atoms.input.cutoff_distance = ip.cutoff_distance
        g.reflect_atoms.input.use_reflection = ip.use_reflection
        g.reflect_atoms.input.total_steps = gp.reflect_atoms.output.total_steps[-1]

        # calc_static
        g.calc_static.input.structure = ip.structure
        g.calc_static.input.project_path = ip.project_path
        g.calc_static.input.job_name = ip.job_name
        g.calc_static.input.positions = gp.reflect_atoms.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # running_average_positions
        g.running_average_pos.input.default.thermalization_steps = ip.thermalization_steps
        g.running_average_pos.input.default.total_steps = ip.total_steps
        g.running_average_pos.input.default.divisor = ip.divisor
        g.running_average_pos.input.default.running_average_positions = ip.running_average_positions

        g.running_average_pos.input.total_steps = gp.running_average_pos.output.total_steps[-1]
        g.running_average_pos.input.divisor = gp.running_average_pos.output.divisor[-1]
        g.running_average_pos.input.running_average_positions = \
            gp.running_average_pos.output.running_average_positions[-1]
        g.running_average_pos.input.positions = gp.reflect_atoms.output.positions[-1]
        g.running_average_pos.input.structure= ip.structure

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'positions': ~gp.reflect_atoms.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.calc_static.output.forces[-1],
            'running_average_positions': ~gp.running_average_pos.output.running_average_positions[-1],
            'divisor': ~gp.running_average_pos.output.divisor[-1],
            'total_steps': ~gp.running_average_pos.output.total_steps[-1],
            'clock': ~gp.clock.output.n_counts[-1]
        }


class FTSEvolutionParallel(FTSEvolution):
    """
    A version of FTSEvolution where the evolution of each image is executed in parallel, thus giving a
        substantial speed-up. Maximum efficiency for parallelization can be achieved by setting the number of cores
        the job can use to the number of images, ie., cores / images = 1. Setting the number of cores greater than
        the number of images gives zero gain, and is wasteful if cores % images != 0.

    Input attributes:
      sleep_time (float): A delay in seconds for database access of results. For sqlite, a non-zero delay maybe
            required. (Default is 0 seconds, no delay.)

    For inherited input and output attributes, refer the `FTSEvolution` protocol.
    """

    def __init__(self, **kwargs):
        super(FTSEvolutionParallel, self).__init__(**kwargs)

        id_ = self.input.default
        # Default values
        # The remainder of the default values are inherited from HarmonicTILD
        id_.sleep_time = 0  # A delay for database access of results. For sqlite, a non-zero delay maybe required.

    def define_vertices(self):
        # Graph components
        g = self.graph
        ip = Pointer(self.input)
        g.create_centroids = CreateJob()
        g.initial_positions = InitialPositions()
        g.initial_forces = Zeros()
        g.initial_velocities = SerialList(RandomVelocity)
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.remove_images = RemoveJob()
        g.create_images = CreateJob()
        g.constrained_evo = ParallelList(_ConstrainedMD, sleep_time=ip.sleep_time)
        g.check_thermalized = IsGEq()
        g.mix = CentroidsRunningAverageMix()
        g.smooth = CentroidsSmoothing()
        g.reparameterize = CentroidsReparameterization()
        g.calc_static_centroids = SerialList(ExternalHamiltonian)
        g.recenter = SerialList(StringRecenter)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.create_centroids,
            g.initial_positions,
            g.initial_forces,
            g.initial_velocities,
            g.cutoff,
            g.check_steps, 'false',
            g.remove_images,
            g.create_images,
            g.constrained_evo,
            g.clock,
            g.check_thermalized, 'true',
            g.mix,
            g.smooth,
            g.reparameterize,
            g.calc_static_centroids,
            g.recenter,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.check_steps, 'false')
        g.starting_vertex = g.create_centroids
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # create_centroids
        g.create_centroids.input.n_images = ip.n_images
        g.create_centroids.input.ref_job_full_path = ip.ref_job_full_path
        g.create_centroids.input.structure = ip.structure_initial

        # initial_positions
        g.initial_positions.input.structure_initial = ip.structure_initial
        g.initial_positions.input.structure_final = ip.structure_final
        g.initial_positions.input.initial_positions = ip.initial_positions
        g.initial_positions.input.n_images = ip.n_images

        # initial_forces
        g.initial_forces.input.shape = ip.structure_initial.positions.shape

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_images
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure_initial.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure_initial
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # remove_images
        g.remove_images.input.default.project_path = ip.project_path
        g.remove_images.input.default.job_names = ip.job_name

        g.remove_images.input.project_path = gp.create_images.output.project_path[-1][-1]
        g.remove_images.input.job_names = gp.create_images.output.job_names[-1]

        # create_images
        g.create_images.input.n_images = ip.n_images
        g.create_images.input.ref_job_full_path = ip.ref_job_full_path
        g.create_images.input.structure = ip.structure_initial

        # constrained_evolution - initiailze
        g.constrained_evo.input.n_children = ip.n_images

        # constrained_evolution - verlet_positions
        g.constrained_evo.direct.structure = ip.structure_initial
        g.constrained_evo.direct.time_step = ip.time_step
        g.constrained_evo.direct.temperature = ip.temperature
        g.constrained_evo.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.constrained_evo.broadcast.default.positions = gp.initial_positions.output.initial_positions[-1]
        g.constrained_evo.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.constrained_evo.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.constrained_evo.broadcast.positions = gp.recenter.output.positions[-1]
        g.constrained_evo.broadcast.velocities = gp.constrained_evo.output.velocities[-1]
        g.constrained_evo.broadcast.forces = gp.recenter.output.forces[-1]

        # constrained_evolution - reflect_string
        g.constrained_evo.direct.default.all_centroid_positions = gp.initial_positions.output.initial_positions[-1]
        g.constrained_evo.broadcast.default.centroid_positions = gp.initial_positions.output.initial_positions[-1]

        g.constrained_evo.direct.all_centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.constrained_evo.broadcast.centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]

        # constrained_evolution - reflect_atoms
        g.constrained_evo.direct.default.total_steps = ip.total_steps
        g.constrained_evo.broadcast.total_steps = gp.constrained_evo.output.total_steps[-1]
        g.constrained_evo.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]

        # constrained_evolution - calc_static
        g.constrained_evo.broadcast.project_path = gp.create_images.output.project_path[-1]
        g.constrained_evo.broadcast.job_name = gp.create_images.output.job_names[-1]

        # constrained_evolution - verlet_velocities
        # takes inputs already specified in verlet_positions

        # constrained_evolution - running_average_positions
        g.constrained_evo.direct.default.thermalization_steps = ip.thermalization_steps
        g.constrained_evo.direct.default.divisor = ip.divisor
        g.constrained_evo.broadcast.default.running_average_positions = \
            gp.initial_positions.output.initial_positions[-1]

        g.constrained_evo.broadcast.divisor = gp.constrained_evo.output.divisor[-1]
        g.constrained_evo.broadcast.running_average_positions = \
            gp.constrained_evo.output.running_average_positions[-1]

        # constrained_evolution - clock
        g.constrained_evo.direct.n_steps = ip.sampling_period

        # clock
        g.clock.input.add_counts = ip.sampling_period

        # check_thermalized
        g.check_thermalized.input.target = gp.constrained_evo.output.total_steps[-1][-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # mix
        g.mix.input.default.centroids_pos_list = gp.initial_positions.output.initial_positions[-1]
        g.mix.input.centroids_pos_list = gp.reparameterize.output.centroids_pos_list[-1]
        g.mix.input.mixing_fraction = ip.mixing_fraction
        g.mix.input.relax_endpoints = ip.relax_endpoints
        g.mix.input.running_average_positions = gp.constrained_evo.output.running_average_positions[-1]
        g.mix.input.structure = ip.structure_initial

        # smooth
        g.smooth.input.kappa = ip.nominal_smoothing
        g.smooth.input.dtau = ip.mixing_fraction
        g.smooth.input.structure = ip.structure_initial
        g.smooth.input.smooth_style = ip.smooth_style
        g.smooth.input.centroids_pos_list = gp.mix.output.centroids_pos_list[-1]

        # reparameterize
        g.reparameterize.input.centroids_pos_list = gp.smooth.output.centroids_pos_list[-1]
        g.reparameterize.input.structure = ip.structure_initial

        # calc_static_centroids
        g.calc_static_centroids.input.n_children = ip.n_images
        g.calc_static_centroids.direct.structure = ip.structure_initial
        g.calc_static_centroids.broadcast.project_path = gp.create_centroids.output.project_path[-1]
        g.calc_static_centroids.broadcast.job_name = gp.create_centroids.output.job_names[-1]
        g.calc_static_centroids.broadcast.positions = gp.reparameterize.output.centroids_pos_list[-1]

        # recenter
        g.recenter.input.n_children = ip.n_images
        g.recenter.direct.default.all_centroid_positions = gp.initial_positions.output.initial_positions[-1]
        g.recenter.broadcast.default.centroid_positions = gp.initial_positions.output.initial_positions[-1]
        g.recenter.broadcast.default.centroid_forces = gp.initial_forces.output.zeros[-1]

        g.recenter.direct.all_centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.recenter.broadcast.centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.recenter.broadcast.centroid_forces = gp.calc_static_centroids.output.forces[-1]
        g.recenter.broadcast.positions = gp.constrained_evo.output.positions[-1]
        g.recenter.broadcast.forces = gp.constrained_evo.output.forces[-1]
        g.recenter.direct.structure = ip.structure_initial

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])


class ProtoFTSEvoPar(Protocol, FTSEvolutionParallel):
    pass



"""
Setuptools based setup module
"""
from setuptools import setup, find_packages
import versioneer


setup(
    name='pyiron_contrib',
    version=versioneer.get_version(),
    description='Repository for user-generated plugins to the pyiron IDE.',
    long_description='http://pyiron.org',

    url='https://github.com/pyiron/pyiron_contrib',
    author='Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department',
    author_email='huber@mpie.de',
    license='BSD',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],

    keywords='pyiron',
    packages=find_packages(exclude=["*tests*"]),
    install_requires=[        
        'matplotlib>=3.4.2',
        'numpy>=1.21.0',
        'pyiron_base>=0.2.17',
        'scipy>=1.7.0',
        'seaborn>=0.11.1'
    ],
    extras_require={
        'atomistic': [
            'ase>=3.22.0',
            'pyiron_atomistics>=0.2.17',
        ],
        'fenics': [
            'fenics>=2019.1.0',
            'mshr>=2019.1.0',
        ],
        'image': ['scikit-image>=0.18.1'],
    },
    cmdclass=versioneer.get_cmdclass(),
    
)



from __future__ import print_function
# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
import inspect
from importlib import import_module
from pyiron_contrib.protocol.generic import LoggerMixin
from weakref import WeakKeyDictionary
from collections import UserList
from operator import itemgetter

"""
Code used by the image library which isn't specific to the task of images, but which doesn't have a home anywhere else
yet.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Feb 3, 2020"


def _elementwise_other(function):
    def wrapper(self, other):
        if hasattr(other, '__len__') and len(other) == len(self):
            return DistributingList(getattr(obj, function.__name__)(oth) for obj, oth in zip(self, other))
        else:
            return DistributingList(getattr(obj, function.__name__)(other) for obj in self)
    return wrapper


def _decorate_by_name(decorator, names):
    def decorate(cls):
        for name in names:
            if hasattr(cls, name):
                setattr(cls, name, decorator(getattr(cls, name)))
        return cls
    return decorate


@_decorate_by_name(decorator=_elementwise_other,
                   names=[
                       '__eq__', '__ne__', '__lt__', '__gt__', '__le__', '__ge__',
                       '__add__', '__sub__', '__mul__', '__floordiv__', '__div__',
                       '__mod__', '__divmod__', '__pow__',
                       '__lshift__', '__rshift__', '__and__', '__or__', '__xor__',
                       '__radd__', '__rsub__', '__rmul__', '__rfloordiv__', '__rdiv__',
                       '__rmod__', '__rdivmod__', '__rpow__',
                       '__rlshift__', '__rrshift__', '__rand__', '__ror__', '__rxor__',
                       '__iadd__', '__isub__', '__imul__', '__ifloordiv__', '__idiv__',
                       '__imod__', '__idivmod__', '__ipow__',
                       '__ilshift__', '__irshift__', '__iand__', '__ior__', '__ixor__',
                   ])
class DistributingList(UserList):
    """
    A list-like class which resolves attribute and function calls by returning a list-like class of the corresponding
    call on each child object.

    TODO:
        - __dir__ for autocomplete?
        - Elementwise decoration for in-place magic methods
        - Setattr stuff
        - Why do strings not have an __iadd__ method?! They clearly work with +=, so what is it calling?
    """

    def __getattr__(self, item):
        return DistributingList([getattr(obj, item) for obj in self])

    def __call__(self, *args, **kwargs):
        ret = [obj.__call__(*args, **kwargs) for obj in self]
        if all(x is None for x in ret):
            return None
        else:
            return DistributingList(ret)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return DistributingList(super(DistributingList, self).__getitem__(item))
        elif isinstance(item, (list, tuple, np.ndarray)):
            if len(self) == len(item) and all(isinstance(i, bool) for i in item):
                return DistributingList(obj for obj, i in zip(self, item) if i)
            else:
                return DistributingList(itemgetter(*item)(self))
        else:
            return super(DistributingList, self).__getitem__(item)


class LockedIfAttributeTrue(LoggerMixin):
    """
    A descriptor which prevents modification when the provided attribute of the owning instance is True. 
    
    Educational credit goes to:
    https://nbviewer.jupyter.org/urls/gist.github.com/ChrisBeaumont/5758381/raw/descriptor_writeup.ipynb
    
    Attributes:
        default: The default value for the descriptor.
        attribute_name (str): The name of the attribute to look for in the owner instance when determining lock state.
        data (weakref.WeakKeyDictionary): A container to track instances.
        name (str): The name of the attribute the descriptor is being assigned to.
    """
    def __init__(self, default, attribute_name):
        self.default = default
        self.attribute_name = attribute_name
        self.data = WeakKeyDictionary()
        self.name = None

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        return self.data.get(instance, self.default)

    def __set__(self, instance, value):
        if getattr(instance, self.attribute_name):
            self.logger.warning("The attribute '{}' cannot be modified while '{}' is True".format(self.name,
                                                                                                  self.attribute_name))
        else:
            self.data[instance] = value


class ModuleScraper:
    """
    A class which scrapes through a module and applies classes and primitives found as attributes of itself, functions
    found as methods of itself, and sub-modules found recursively as new `ModuleScraper` attributes of itself.

    A decorator can optionally be applied to all functions found. This is the real strength, since

    Note:
        Doesn't do anything until its `activate` method is called.

    TODO:
        - Allow decorators also for classes
        - Perhaps some sort of more complex mapping so that the decorators aren't just uniformly applied to every
          function/class that's found, but can be selectively applied.

    Attributes:
        safe (bool): Whether to skip values beginning with an underscore. (Default is True, do skip.)
        recursive (bool): Whether to recursively activate submodules. (Default is False: submodules are noted by
            creating a new `ModuleScraper` instance and setting it as an attribute, but contents cannot be accessed
            until the submodule itself is explicitly activated.)
        scrape_functions (bool): Whether to look for functions in the module. (Default is True.)
        scrape_classes (bool): Whether to look for class definitions in the module. (Default is True.)
        scrape_primitives (bool): Whether to look for primitives in the module. (Default is True.)
        primitives_list (tuple/list): A list of which types count as primitive. (Default is None, which uses `(int,
            float, bool, numpy.ndarray)`.)
    """

    safe = LockedIfAttributeTrue(True, '_activated')
    recursive = LockedIfAttributeTrue(True, '_activated')
    scrape_functions = LockedIfAttributeTrue(True, '_activated')
    scrape_classes = LockedIfAttributeTrue(True, '_activated')
    scrape_primitives = LockedIfAttributeTrue(True, '_activated')
    primitives_list = LockedIfAttributeTrue((int, float, bool, np.ndarray), '_activated')

    def __init__(
            self,
            module,
            decorator=None,
            decorator_args=None,
            safe=True,
            recursive=True,
            scrape_functions=True,
            scrape_classes=True,
            scrape_primitives=True,
            primitives_list=None
    ):
        """
        Args:
            module (module/str): The module from which to scrape, or the name of the module from which to escape, e.g.
                `skimage.filters` or `'skimage.filters'`.
            decorator (fnc): A decorator function to apply to scraped functions. (Default is None, no decorator.)
            decorator_args (tuple/list): Arguments to pass to the decorator. (Default is None, no args.)
            safe (bool): Whether to skip values beginning with an underscore. (Default is True, do skip.)
            recursive (bool): Whether to recursively activate submodules. (Default is False: submodules are noted by
                creating a new `ModuleScraper` instance and setting it as an attribute, but contents cannot be accessed
                until the submodule itself is explicitly activated.)
            scrape_functions (bool): Whether to look for functions in the module. (Default is True.)
            scrape_classes (bool): Whether to look for class definitions in the module. (Default is True.)
            scrape_primitives (bool): Whether to look for primitives in the module. (Default is True.)
            primitives_list (tuple/list): A list of which types count as primitive. (Default is None, which uses `(int,
                float, bool, numpy.ndarray)`.)
        """
        self._module = module
        self._decorator = decorator
        self._decorator_args = decorator_args or ()
        self._activated = False

        self.safe = safe
        self.recursive = recursive
        self.scrape_functions = scrape_functions
        self.scrape_classes = scrape_classes
        self.scrape_primitives = scrape_primitives
        self.primitives_list = primitives_list or (int, float, bool, np.ndarray)

    def activate(self):
        """
        Parse the module assigned at instantiation, using decorator information assigned at instantiation.
        """
        if inspect.ismodule(self._module):
            module = self._module
        else:
            module = import_module(self._module)

        for name, obj in inspect.getmembers(module):
            if self.safe and name[0] == '_':
                continue
            elif self.recursive and inspect.ismodule(obj) and obj.__package__ == module.__package__:
                # Behave recursively for submodules
                submodule = ModuleScraper(
                    obj,
                    decorator=self._decorator,
                    decorator_args=self._decorator_args,
                    safe=self.safe,
                    recursive=self.recursive,
                    scrape_functions=self.scrape_functions,
                    scrape_classes=self.scrape_classes,
                    scrape_primitives=self.scrape_primitives,
                    primitives_list=self.primitives_list
                )
                setattr(self, name, submodule)
                if self.recursive:
                    submodule.activate()
            elif self.scrape_functions and inspect.isfunction(obj):
                # Set all module functions as methods
                if self._decorator is not None:
                    fnc = self._decorator(*self._decorator_args)(obj)
                else:
                    fnc = obj
                setattr(self, name, fnc)
            elif self.scrape_classes and inspect.isclass(obj):
                # Grab classes
                setattr(self, name, obj)
            elif self.scrape_primitives and isinstance(obj, self.primitives_list):
                # Grab primitives
                setattr(self, name, obj)

        self._activated = True

    def __getattr__(self, item):
        if not self._activated:
            self.activate()
            return getattr(self, item)
        else:
            if inspect.ismodule(self._module):
                name = self._module.__name__.split('.')[-1]
            else:
                name = self._module.split('.')[-1]
            raise AttributeError(
                "'{0}' has no attribute '{1}' first.".format(name, item))

    def to_hdf(self):
        pass

    def from_hdf(self):
        pass



from __future__ import print_function
# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Jan 30, 2020"


def brightness_filter(image, sigma=10, bins=100, deg=10, plot=False):
    """
    Automatically detect a threshold between 'dark' and 'light' pixel values by looking for a minima in the pixel value
    histogram, fitting a polynomial, and finding the trough between the two highest peaks.

    Just exists as a near-trivial example

    Args:
        image (image.Image): The image to filter.
        sigma (float): The amount of gaussian smearing to apply to the image data before binning brightness. (Default is
            10.)
        bins (int): How many bins to put the pixel brightness into. (Default is 100)
        deg (int): Degree of polynomial to fit to the binned brightness data. (Default is 10.)
        plot (bool): Whether to plot a summary of the filtering process. (Default is False.)

    Returns:
        (float): The fraction of pixels darker than the threshold.
        (float): The threshold used.
        (numpy.ndarray): The mask of values from the smoothed image below the threshold.
    """
    # Smooth out to avoid checker-board of secondary phase
    image.reload_data()
    raw_data = image.data.copy()
    image.filters.gaussian(sigma=sigma)
    signal = image.data.flatten()

    # Find 'dark' and 'light' signal peaks
    counts, edges = np.histogram(image.data.flatten(), bins=bins)
    bincenters = 0.5 * (edges[:-1] + edges[1:])
    coeffs = np.polyfit(bincenters, counts, deg=deg)
    poly = np.poly1d(coeffs)
    extrema = poly.deriv().r[1:-1]
    extremal_bin_ids = np.array([np.argmin(np.abs((bincenters - x))) for x in extrema], dtype=int)

    # Find the trough location between the two biggest peaks
    ultimate_id, penultimate_id = np.argsort(poly(bincenters[extremal_bin_ids]))[-2:]
    trough_id = extremal_bin_ids[int(0.5 * (ultimate_id + penultimate_id))]
    peak_id, pen_peak_id = extremal_bin_ids[ultimate_id], extremal_bin_ids[penultimate_id]
    if not (bincenters[peak_id] < bincenters[trough_id] < bincenters[pen_peak_id]) and \
            not (bincenters[peak_id] > bincenters[trough_id] > bincenters[pen_peak_id]):
        print("WARNING: Had trouble finding a dividing trough for {}".format(image.source))
    threshold_signal = bincenters[trough_id]

    # Separate phases based on lightness/darkness
    phase_mask = image.data < threshold_signal
    phase_fraction = np.mean(phase_mask)

    if plot:
        # Plot the signal distribution
        _, ax = plt.subplots(figsize=(12, 6))
        sns.distplot(signal, ax=ax, norm_hist=False, kde=False, label='brightness distribution')
        ax.plot(bincenters, poly(bincenters), label='fit')

        ax.axvline(bincenters[peak_id], color='w', linestyle='--')
        ax.axvline(bincenters[trough_id], color='w', label='threshold')
        ax.axvline(bincenters[pen_peak_id], color='w', linestyle='--')
        plt.legend()
        plt.xlabel('Brightness of smoothed image')
        plt.ylabel('Pixel count')
        plt.show()

        # And the resulting map
        _, axes = plt.subplots(ncols=3, figsize=(12, 6))
        axes[0].imshow(raw_data)
        axes[0].set_title('Original')
        axes[1].imshow(image.data)
        axes[1].set_title('Smoothed')
        axes[2].imshow(phase_mask)
        axes[2].set_title('Phase mask')
        plt.show()

    return phase_fraction, threshold_signal, phase_mask


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

from pyiron_contrib.protocol.generic import PrimitiveVertex, CompoundVertex, Protocol
from pyiron_contrib.protocol.utils import ensure_iterable
from pyiron_contrib.protocol.primitive.one_state import ExternalHamiltonian, Counter, Norm, Max, GradientDescent
from pyiron_contrib.protocol.primitive.two_state import IsGEq, IsLEq
from pyiron_contrib.protocol.utils import Pointer, IODictionary
import numpy as np

"""
Protocol for running the force-based quantum mechanics/molecular mechanics concurrent coupling scheme described in 
Huber et al., Comp. Mat. Sci. 118 (2016) 259-268
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "June 6, 2019"


class QMMM(CompoundVertex):
    """
    Relax a QM/MM coupled system.

    Needs an MM job (incl. superstructure), a QM job reference, QM target site ids, optional target site chemistry,
    optional domain ids, optional QM cell instructions?

    During setup, the domain indices are checked -- if None, build them from scratch using shells and the cell...stuff?

    Input attributes:
        structure (pyiron.atomistics.structure.atoms.Atoms): The full region I+II superstructure.
        mm_ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies of the MM domains.
        qm_ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies of the QM domain.
        domain_ids (dict): A dictionary of ids, `seed`, `core`, `buffer`, and `filler`, for mapping atoms of the MM
            I+II superstructure to the various domains of the QM/MM coupling scheme. (Default is None, gets constructed
            using consecutive shells).
        seed_ids (int/list): The integer id (or ids) of atom(s) in the MM I+II superstructure to base the QM region on.
            These are the only atoms whose species can be changed. (Default is None, which raises an error unless
            `domain_ids` was explicitly provided -- else this parameter must be provided.)
        shell_cutoff (float): Maximum distance for two atoms to be considered neighbours in the construction of shells
            for automatic system partitioning. (Default is None, which will raise an error -- this parameter must be
            provided if domain partitioning is done using `seed_ids` and shells. When `domain_ids` are explicitly
            provided, `shell_cutoff` is not needed.)
        n_core_shells (int): How many neighbour shells around the seed(s) to relax using QM forces. (Default is 2.)
        n_buffer_shells (int): How many neighbour shells around the region I core to relax using MM forces. (Default is
            2.)
        seed_species (str/list): The species for each 'seed' atom in the QM domain. Value(s) should be a atomic symbol,
            e.g. 'Mg'. (Default is None, which leaves all seeds the same as they occur in the input structure.)
        vacuum_width (float): Minimum vacuum distance between atoms in two periodic images of the QM domain. Influences
            the final simulation cell for the QM calculation. (Default is 2.)
        filler_width (float): Length added to the bounding box of the region I atoms to create a new box from which to
            draw filler atoms using the initial MM I+II superstructure. Influences the final simulation cell for the QM
            calculation.If the value is not positive, no filler atoms are used. The second, larger box uses the same
            center as the bounding box of the QM region I, so this length is split equally to the positive and negative
            sides for each cartesian direction. (Default is 6.)
        n_steps (int): The maximum number of minimization steps to make. (Default is 100.)
        f_tol (float): The maximum force on any atom below which the calculation terminates. Only atoms which could be
            relaxed are considered, i.e. QM forces for region I core, and MM forces for region II and I buffer. (Filler
            atoms are not real, so we never care about their forces.) (Default is 1e-4 eV/angstrom.)

    Output attributes:
        energy_mm (float): The total energy in eV of the region I+II superstructure (without seed species) using the MM
            representation.
        energy_qm (float): The total energy in eV of the region I+filler structure (with seed species) using QM
            representation.
        energy_mm_small (float): The total energy in eV of the region I+filler structure (without seed species) using
            the representation.
        energy_qmmm (float): The composite QM/MM energy: `energy_mm` + `energy_qm` - `energy_mm_small` in eV.
        max_force (float): The largest atomic force among QM forces on region I core atoms or MM forces on region I
            buffer and region II atoms.
        positions (numpy.ndarray): The per-atom vector of cartesian positions for the entire region I+II superstructure.
    """

    DefaultWhitelist = {
        'calc_static_mm': {
            'output': {
                'energy_pot': 1,
            },
        },
        'calc_static_qm': {
            'output': {
                'energy_pot': 1,
            },
        },
    }

    def __init__(self, **kwargs):
        super(QMMM, self).__init__(**kwargs)

        id_ = self.input.default
        id_.domain_ids = None
        id_.seed_ids = None
        id_.shell_cutoff = None
        id_.n_core_shells = 2
        id_.n_buffer_shells = 2
        id_.vacuum_width = 2.
        id_.filler_width = 6.

        id_.n_steps = 100
        id_.f_tol = 1e-4

        id_.gamma0 = 0.1
        id_.fix_com = True
        id_.use_adagrad = False

    def define_vertices(self):
        # Components
        g = self.graph
        g.partition = PartitionStructure()
        g.calc_static_mm = ExternalHamiltonian()
        g.calc_static_qm = ExternalHamiltonian()
        g.clock = Counter()
        g.force_norm_mm = Norm()
        g.force_norm_qm = Norm()
        g.max_force_mm = Max()
        g.max_force_qm = Max()
        g.check_force_mm = IsLEq()
        g.check_force_qm = IsLEq()
        g.check_steps = IsGEq()
        g.update_buffer_qm = AddDisplacements()
        g.update_core_mm = AddDisplacements()
        g.gradient_descent_mm = GradientDescent()
        g.gradient_descent_qm = GradientDescent()
        g.calc_static_small = ExternalHamiltonian()

    def define_execution_flow(self):
        g = self.graph
        g.make_pipeline(
            g.partition,
            g.check_steps, 'false',
            g.calc_static_mm,
            g.calc_static_qm,
            g.force_norm_mm,
            g.max_force_mm,
            g.check_force_mm, 'true',
            g.force_norm_qm,
            g.max_force_qm,
            g.check_force_qm, 'false',
            g.gradient_descent_mm,
            g.gradient_descent_qm,
            g.update_buffer_qm,
            g.update_core_mm,
            g.clock,
            g.check_steps
        )
        g.make_edge(g.check_force_mm, g.gradient_descent_mm, 'false')
        g.make_edge(g.check_force_qm, g.calc_static_small, 'true')
        g.make_edge(g.check_steps, g.calc_static_small, 'true')
        g.starting_vertex = g.partition
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        gp = Pointer(self.graph)
        ip = Pointer(self.input)
        g = self.graph

        g.partition.input.structure = ip.structure
        g.partition.input.domain_ids = ip.domain_ids
        g.partition.input.seed_ids = ip.seed_ids
        g.partition.input.shell_cutoff = ip.shell_cutoff
        g.partition.input.n_core_shells = ip.n_core_shells
        g.partition.input.n_buffer_shells = ip.n_buffer_shells
        g.partition.input.vacuum_width = ip.vacuum_width
        g.partition.input.filler_width = ip.filler_width
        g.partition.input.seed_species = ip.seed_species

        g.calc_static_mm.input.ref_job_full_path = ip.mm_ref_job_full_path
        g.calc_static_mm.input.structure = gp.partition.output.mm_full_structure[-1]
        g.calc_static_mm.input.default.positions = gp.partition.output.mm_full_structure[-1].positions
        g.calc_static_mm.input.positions = gp.update_core_mm.output.positions[-1]

        g.calc_static_small.input.ref_job_full_path = ip.mm_ref_job_full_path
        g.calc_static_small.input.structure = gp.partition.output.mm_small_structure[-1]
        g.calc_static_small.input.default.positions = gp.partition.output.mm_small_structure[-1].positions
        g.calc_static_small.input.positions = gp.update_buffer_qm.output.positions[-1]

        g.calc_static_qm.input.ref_job_full_path = ip.qm_ref_job_full_path
        g.calc_static_qm.input.structure = gp.partition.output.qm_structure[-1]
        g.calc_static_qm.input.default.positions = gp.partition.output.qm_structure[-1].positions
        g.calc_static_qm.input.positions = gp.update_buffer_qm.output.positions[-1]

        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        g.force_norm_mm.input.x = gp.calc_static_mm.output.forces[-1][
            gp.partition.output.domain_ids[-1]['except_core']
        ]
        g.force_norm_mm.input.ord = 2
        g.force_norm_mm.input.axis = -1

        g.max_force_mm.input.a = gp.force_norm_mm.output.n[-1]
        g.check_force_mm.input.target = gp.max_force_mm.output.amax[-1]
        g.check_force_mm.input.threshold = ip.f_tol

        g.force_norm_qm.input.x = gp.calc_static_qm.output.forces[-1][
            gp.partition.output.domain_ids_qm[-1]['only_core']
        ]
        g.force_norm_qm.input.ord = 2
        g.force_norm_qm.input.axis = -1

        g.max_force_qm.input.a = gp.force_norm_qm.output.n[-1]
        g.check_force_qm.input.target = gp.max_force_qm.output.amax[-1]
        g.check_force_qm.input.threshold = ip.f_tol

        g.gradient_descent_mm.input.forces = gp.calc_static_mm.output.forces[-1]
        g.gradient_descent_mm.input.default.positions = gp.partition.output.mm_full_structure[-1].positions
        g.gradient_descent_mm.input.positions = gp.update_core_mm.output.positions[-1]
        g.gradient_descent_mm.input.masses = gp.partition.output.mm_full_structure[-1].get_masses
        g.gradient_descent_mm.input.mask = gp.partition.output.domain_ids[-1]['except_core']
        g.gradient_descent_mm.input.gamma0 = ip.gamma0
        g.gradient_descent_mm.input.fix_com = ip.fix_com
        g.gradient_descent_mm.input.use_adagrad = ip.use_adagrad

        g.gradient_descent_qm.input.forces = gp.calc_static_qm.output.forces[-1]
        g.gradient_descent_qm.input.default.positions = gp.partition.output.qm_structure[-1].positions
        g.gradient_descent_qm.input.positions = gp.update_buffer_qm.output.positions[-1]
        g.gradient_descent_qm.input.masses = gp.partition.output.qm_structure[-1].get_masses
        g.gradient_descent_qm.input.mask = gp.partition.output.domain_ids_qm[-1]['only_core']
        g.gradient_descent_qm.input.gamma0 = ip.gamma0
        g.gradient_descent_qm.input.fix_com = ip.fix_com
        g.gradient_descent_qm.input.use_adagrad = ip.use_adagrad

        g.update_core_mm.input.default.target = gp.partition.output.mm_full_structure[-1].positions
        g.update_core_mm.input.target = gp.gradient_descent_mm.output.positions[-1]
        g.update_core_mm.input.target_mask = [
            gp.partition.output.domain_ids[-1]['seed'],
            gp.partition.output.domain_ids[-1]['core']
        ]
        g.update_core_mm.input.displacement = gp.gradient_descent_qm.output.displacements[-1]
        g.update_core_mm.input.displacement_mask = [
            gp.partition.output.domain_ids_qm[-1]['seed'],
            gp.partition.output.domain_ids_qm[-1]['core']
        ]

        g.update_buffer_qm.input.default.target = gp.partition.output.qm_structure[-1].positions
        g.update_buffer_qm.input.target = gp.gradient_descent_qm.output.positions[-1]
        g.update_buffer_qm.input.target_mask = gp.partition.output.domain_ids_qm[-1]['buffer']
        g.update_buffer_qm.input.displacement = gp.gradient_descent_mm.output.displacements[-1]
        g.update_buffer_qm.input.displacement_mask = gp.partition.output.domain_ids[-1]['buffer']

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        e_mm = ~gp.calc_static_mm.output.energy_pot[-1]
        e_qm = ~gp.calc_static_qm.output.energy_pot[-1]
        e_mm_small = ~gp.calc_static_small.output.energy_pot[-1]
        try:  # If we terminate by step count, the QM force checker may never have gotten called
            max_force = max(~gp.max_force_mm.output.amax[-1], ~gp.max_force_qm.output.amax[-1])
        except KeyError:
            max_force = ~gp.max_force_mm.output.amax[-1]
        try:  # We might also be converged before ever running a step (e.g. for perfect bulk)
            positions = ~gp.update_core_mm.output.positions[-1]
        except KeyError:
            positions = self.input.structure.positions
        return {
            'energy_mm': e_mm,
            'energy_qm': e_qm,
            'energy_mm_small': e_mm_small,
            'energy_qmmm': e_mm + e_qm - e_mm_small,
            'max_force': max_force,
            'positions': positions
        }

    def show_mm(self):
        try:
            mm_full_structure = self.graph.partition.output.mm_full_structure[-1]
            mm_full_structure.positions = self.graph.update_core_mm.output.positions[-1]
            domain_ids = self.graph.partition.output.domain_ids[-1]
        except KeyError:
            partitioned = self.partition_input()
            mm_full_structure = partitioned['mm_full_structure']
            domain_ids = partitioned['domain_ids']

        color = 4 * np.ones(len(mm_full_structure))  # This 5th colour makes the balance of atoms
        for n, group in enumerate(['seed', 'core', 'buffer', 'filler']):
            color[domain_ids[group]] = n

        return mm_full_structure.plot3d(scalar_field=color)

    def show_qm(self):
        try:
            qm_structure = self.graph.partition.output.qm_structure[-1]
            qm_structure.positions = self.graph.update_buffer_qm.output.positions[-1]
            domain_ids_qm = self.graph.partition.output.domain_ids_qm[-1]
        except KeyError:
            partitioned = self.partition_input()
            qm_structure = partitioned['qm_structure']
            domain_ids_qm = partitioned['domain_ids_qm']

        color = 4 * np.ones(len(qm_structure))  # If you see this 5th colour, something is wrong
        for n, group in enumerate(['seed', 'core', 'buffer', 'filler']):
            color[domain_ids_qm[group]] = n

        return qm_structure.plot3d(scalar_field=color)

    def show_boxes(self):
        try:
            structure = self.graph.partition.output.structure[-1]
            qm_structure = self.graph.partition.output.qm_structure[-1]
        except KeyError:
            partitioned = self.partition_input()
            structure = partitioned['mm_full_structure']
            qm_structure = partitioned['qm_structure']
        self._plot_boxes([structure.cell, qm_structure.cell],
                         colors=['r', 'b'],
                         titles=['MM Superstructure', 'QM Structure'])

    def partition_input(self):
        i = self.input
        partition = PartitionStructure()
        return partition.command(
            i.structure,
            i.domain_ids,
            i.seed_ids, i.shell_cutoff, i.n_core_shells, i.n_buffer_shells,
            i.vacuum_width, i.filler_width,
            i.seed_species
        )

    @staticmethod
    def _plot_boxes(cells, translate=None, colors=None, titles=None, default_color='b', size=(29, 21)):
        """
        Plots one or a list of cells in xy, yz and xt projection
        Args:
            cells (numpy.ndarray/list): The cells to plot. A list of (3,3) or a single (3,3) matrix
            translate (list): list of translations vectors for each cell list of (3,) numpy arrays.
            colors (list): list of colors for each cell in the list. e.g ['b', 'g', 'y'] or 'bgky'
            titles (list): list of names displayed for each cell. list of str
            default_color (str): if colors is not specified this color will be applied to all cells
            size (float,float): A tuple of two float specifiyng the size of the plot in centimeters

        Returns:
            matplotlib.figure: The figure which will be displayed
        """
        import matplotlib.pyplot as plt
        from matplotlib.lines import Line2D

        # Make sure that looping over this properties works smoothly
        cells = ensure_iterable(cells)
        translate = translate or [np.zeros(3,)]*len(cells)
        colors = colors or [default_color]*len(cells)
        titles = ensure_iterable(titles)

        # Scaled coordinates for all edges of the cell
        edges = [
            [(0, 0, 0), (1, 0, 0)],
            [(0, 0, 0), (0, 1, 0)],
            [(0, 0, 0), (0, 0, 1)],
            [(1, 1, 0), (1, 0, 0)],
            [(1, 1, 0), (0, 1, 0)],
            [(1, 1, 0), (1, 1, 1)],
            [(1, 0, 1), (1, 0, 0)],
            [(1, 0, 1), (1, 1, 1)],
            [(1, 0, 1), (0, 0, 1)],
            [(0, 1, 1), (1, 1, 1)],
            [(0, 1, 1), (0, 0, 1)],
            [(0, 1, 1), (0, 1, 0)]
        ]

        # Map axis to indices
        axis_mapping = {a: i for a, i in zip('xyz', range(3))}
        # Planes to plot
        planes = ['xy', 'yz', 'xz']
        # Get subplots and set the size in inches
        fig, axes = plt.subplots(1, len(planes))
        fig.set_size_inches(*[s / 2.54 for s in size])

        for i, plane in enumerate(planes):
            axis = axes[i]
            axis.set_aspect('equal')
            axis.title.set_text('{} plane'.format(plane))
            indices = [axis_mapping[a] for a in plane]
            for j, (cell, color, translation) in enumerate(zip(cells, colors, translate)):
                calc = lambda mul_, cell_: sum([m_ * vec_ for m_, vec_ in zip(mul_, cell_)])
                coords = [(calc(start, cell) + translation, calc(end, cell) + translation) for start, end in edges]
                for start, end in coords:
                    sx, sy = start[indices]
                    ex, ey = end[indices]
                    axis.plot([sx, ex], [sy, ey], color=color)
        # Create dummy legend outside
        legend_lines = [Line2D([0], [0], color=col or default_color, lw=2) for col in colors]
        legend_titles = [title or 'Box {}'.format(i + 1) for i, title in enumerate(titles)]
        plt.figlegend(legend_lines, legend_titles, loc='lower center', fancybox=True, shadow=True)
        return plt


class AddDisplacements(PrimitiveVertex):

    def __init__(self, name=None):
        super(AddDisplacements, self).__init__(name=name)
        self.input.default.target_mask = None
        self.input.default.displacement_mask = None

    def command(self, target, displacement, target_mask, displacement_mask):
        result = target.copy()
        if target_mask is not None and isinstance(target_mask, list):
            target_mask = np.concatenate(target_mask)
        if displacement_mask is not None and isinstance(displacement_mask, list):
            displacement_mask = np.concatenate(displacement_mask)
        if target_mask is not None and displacement_mask is not None:
            result[target_mask] += displacement[displacement_mask]
        elif target_mask is not None and displacement_mask is None:
            result[target_mask] += displacement
        elif target_mask is None and displacement_mask is not None:
            result += displacement[displacement_mask]
        else:
            result += displacement
        return {
            'positions': result
        }


class PartitionStructure(PrimitiveVertex):
    """

    """
    def __init__(self, name=None):
        super(PartitionStructure, self).__init__(name=name)
        id_ = self.input.default
        id_.domain_ids = None
        id_.seed_ids = None
        id_.shell_cutoff = None
        id_.n_core_shells = None
        id_.n_buffer_shells = None
        id_.filler_width = None

    def command(
            self,
            structure,
            domain_ids,
            seed_ids, shell_cutoff, n_core_shells, n_buffer_shells,
            vacuum_width, filler_width,
            seed_species
    ):
        domain_ids, domain_ids_qm, mm_small_structure = self._set_qm_structure(
            structure,
            domain_ids,
            seed_ids, shell_cutoff, n_core_shells, n_buffer_shells,
            vacuum_width, filler_width
        )
        qm_structure = self._change_qm_species(mm_small_structure, domain_ids_qm, seed_species)
        domain_ids_qm['only_core'] = self._only_core(domain_ids_qm)
        domain_ids['except_core'] = self._except_core(structure, domain_ids)
        return {
            'mm_full_structure': structure.copy(),
            'mm_small_structure': mm_small_structure.copy(),
            'qm_structure': qm_structure.copy(),
            'domain_ids': domain_ids,
            'domain_ids_qm': domain_ids_qm
        }

    def _set_qm_structure(
            self,
            superstructure,
            domain_ids,
            seed_ids, shell_cutoff, n_core_shells, n_buffer_shells,
            vacuum_width, filler_width
    ):
        if domain_ids is not None and seed_ids is not None:
            raise ValueError('Only *one* of `seed_ids` and `domain_ids` may be provided.')
        elif domain_ids is not None:
            seed_ids = domain_ids['seed']
            core_ids = domain_ids['core']
            buffer_ids = domain_ids['buffer']
            filler_ids = domain_ids['filler']
            region_I_ids = np.concatenate((seed_ids, core_ids, buffer_ids))
            bb = self._get_bounding_box(superstructure[np.concatenate((region_I_ids, filler_ids))])
        elif seed_ids is not None:
            shells = self._build_shells(superstructure, n_core_shells + n_buffer_shells, shell_cutoff, seed_ids)
            core_ids = np.concatenate(shells[:n_core_shells])
            buffer_ids = np.concatenate(shells[n_core_shells:])
            region_I_ids = np.concatenate((seed_ids, core_ids, buffer_ids))
            bb = self._get_bounding_box(superstructure[region_I_ids])
            extra_box = 0.5 * np.array(filler_width)
            bb[:, 0] -= extra_box
            bb[:, 1] += extra_box

            # Store it because get bounding box return a tight box and is different
            filler_ids = self._get_ids_within_box(superstructure, bb)
            filler_ids = np.setdiff1d(filler_ids, region_I_ids)

            bb = self._get_bounding_box(superstructure[np.concatenate((region_I_ids, filler_ids))])

            domain_ids = {'seed': seed_ids, 'core': core_ids, 'buffer': buffer_ids, 'filler': filler_ids}
        else:
            raise ValueError('At least *one* of `seed_ids` and `domain_ids` must be provided.')

        # Build the domain ids in the qm structure
        qm_structure = None
        domain_ids_qm = {}
        offset = 0
        for key, ids in domain_ids.items():
            if qm_structure is None:
                qm_structure = superstructure[ids]
            else:
                qm_structure += superstructure[ids]
            id_length = len(ids)
            domain_ids_qm[key] = np.arange(id_length) + offset
            offset += id_length

        # And put everything in a box near (0,0,0)
        extra_vacuum = 0.5 * vacuum_width
        bb[:, 0] -= extra_vacuum
        bb[:, 1] += extra_vacuum

        # If the bounding box is larger than the MM superstructure
        bs = np.abs(bb[:, 1] - bb[:, 0])
        supercell_lengths = [np.linalg.norm(row) for row in superstructure.cell]
        shrinkage = np.array([(box_size - cell_size) / 2.0 if box_size > cell_size else 0.0
                              for box_size, cell_size in zip(bs, supercell_lengths)])
        if np.any(shrinkage > 0):
            self.logger.info('The QM box is larger than the MM Box therefore I\'ll shrink it')
            bb[:, 0] += shrinkage
            bb[:, 1] -= shrinkage
        elif any([0.9 < box_size / cell_size < 1.0 for box_size, cell_size in zip(bs, supercell_lengths)]):
            # Check if the box is just slightly smaller than the superstructure cell
            self.logger.warn(
                'Your cell is nearly as large as your supercell. Probably you want to expand it a little bit')

        qm_structure.cell = np.identity(3) * np.ptp(bb, axis=1)

        box_center = tuple(np.dot(np.linalg.inv(superstructure.cell), np.mean(bb, axis=1)))
        qm_structure.wrap(center=box_center)
        # Wrap it to the unit cell
        qm_structure.positions = np.dot(qm_structure.get_scaled_positions(), qm_structure.cell)

        return domain_ids, domain_ids_qm, qm_structure

    @staticmethod
    def _change_qm_species(qm_structure, domain_ids_qm, seed_species):
        qm_structure_alchemized = qm_structure.copy()
        for index, species in zip(domain_ids_qm['seed'], seed_species):
            qm_structure_alchemized[index] = species
        return qm_structure_alchemized

    @staticmethod
    def _get_ids_within_box(structure, box):
        """
        Finds all the atoms in a structure who have a periodic image inside the bounding box.

        Args:
            structure (Atoms): The structure to search.
            box (np.ndarray): A 3x2 array of the x-min through z-max values for the bounding rectangular prism.

        Returns:
            np.ndarray: The integer ids of the atoms inside the box.
        """
        box_center = np.mean(box, axis=1)
        box_center_direct = np.dot(np.linalg.inv(structure.cell), box_center)
        # Wrap atoms so that they are the closest image to the box center
        wrapped_structure = structure.copy()
        wrapped_structure.wrap(center=tuple(box_center_direct))
        pos = wrapped_structure.positions
        # Keep only atoms inside the box limits
        masks = []
        for d in np.arange(len(box)):
            masks.append(pos[:, d] > box[d, 0])
            masks.append(pos[:, d] < box[d, 1])
        total_mask = np.prod(masks, axis=0).astype(bool)
        return np.arange(len(structure), dtype=int)[total_mask]

    @staticmethod
    def _build_shells(structure, n_shells, shell_cutoff, seed_ids):
        indices = [seed_ids]
        current_shell_ids = seed_ids
        for _ in range(n_shells):
            neighbors = structure.get_neighbors(id_list=current_shell_ids, cutoff=shell_cutoff)
            new_ids = np.setdiff1d(
                np.unique(np.concatenate(neighbors.indices)),
                np.concatenate(indices)
            )
            indices.append(new_ids)
            current_shell_ids = new_ids
        # Pop seed ids
        indices.pop(0)
        return indices

    @staticmethod
    def _get_bounding_box(structure):
        """
        Finds the smallest rectangular prism which encloses all atoms in the structure after accounting for periodic
        boundary conditions.

        So what's the problem?

        |      ooooo  |, easy, wrap by CoM (centre of mass)
        |ooo        oo|, easy, wrap by CoM
        |o    ooo    o|, uh-oh! Need this function.


        Args:
            structure (Atoms): The structure to bound.

        Returns:
            numpy.ndarray: A 3x2 array of the x-min through z-max values for the bounding rectangular prism.
        """
        wrapped_structure = structure.copy()
        # Take the frist positions and wrap the atoms around there to determine the size of the bounding box
        wrap_center = tuple(np.dot(np.linalg.inv(structure.cell), structure.positions[0, :]))
        wrapped_structure.wrap(center=wrap_center)

        bounding_box = np.vstack([
            np.amin(wrapped_structure.positions, axis=0),
            np.amax(wrapped_structure.positions, axis=0)
        ]).T
        return bounding_box

    @staticmethod
    def _except_core(structure, domain_ids):
        seed_ids = np.array(domain_ids['seed'])
        core_ids = np.array(domain_ids['core'])
        return np.setdiff1d(np.arange(len(structure)), np.concatenate([seed_ids, core_ids]))

    @staticmethod
    def _only_core(domain_ids_qm):
        return np.concatenate([domain_ids_qm['seed'], domain_ids_qm['core']])


class ProtocolQMMM(Protocol, QMMM):
    pass
